package de.hft.stuttgart.citydoctor2.gui;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.scene.control.Alert;
import javafx.scene.control.CheckBox;
import javafx.scene.control.ProgressIndicator;
import javafx.scene.control.Tooltip;
import javafx.stage.DirectoryChooser;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.hft.stuttgart.citydoctor2.check.DefaultParameter;
import de.hft.stuttgart.citydoctor2.check.FilterConfiguration;
import de.hft.stuttgart.citydoctor2.check.GlobalParameters;
import de.hft.stuttgart.citydoctor2.check.Requirement;
import de.hft.stuttgart.citydoctor2.check.RequirementConfiguration;
import de.hft.stuttgart.citydoctor2.check.RequirementType;
import de.hft.stuttgart.citydoctor2.check.Unit;
import de.hft.stuttgart.citydoctor2.check.ValidationConfiguration;
import de.hft.stuttgart.citydoctor2.checks.Checks;
import de.hft.stuttgart.citydoctor2.gui.tree.TreeRequirement;
import de.hft.stuttgart.citydoctor2.utils.Localization;
import javafx.application.Platform;
import javafx.event.Event;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.control.ProgressBar;
import javafx.scene.control.ScrollPane;
import javafx.scene.control.Tab;
import javafx.scene.control.TabPane;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableView;
import javafx.scene.control.TextField;
import javafx.scene.control.TreeItem;
import javafx.scene.control.TreeTableColumn;
import javafx.scene.control.TreeTableView;
import javafx.scene.control.cell.CheckBoxTreeTableCell;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.control.cell.TextFieldTableCell;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.VBox;
import javafx.stage.FileChooser;
import javafx.stage.FileChooser.ExtensionFilter;
import javafx.stage.Modality;
import javafx.stage.Stage;
import javafx.stage.Window;
import javafx.util.StringConverter;
import javafx.util.converter.DefaultStringConverter;

public class CheckDialog {

	private static final String NUMBER_OF_ROUNDING_PLACES = "numberOfRoundingPlaces";

	private static final Logger logger = LogManager.getLogger(CheckDialog.class);

	private static final String CHECK_ENABLED = Localization.getText("CheckDialog.checkEnabled");
	private static final String NAME = Localization.getText("CheckDialog.checkName");
	private static final String VALUE = Localization.getText("CheckDialog.parameterValue");
	private static final String UNIT = Localization.getText("CheckDialog.parameterUnit");

	private Stage stage;

	@FXML
	private TreeTableView<TreeRequirement> geometricTable;

	@FXML
	private TreeTableView<TreeRequirement> semanticTable;

	@FXML
	private TableView<GlobalParameter> globalParametersTable;

	@FXML
	private ScrollPane scrollPane;

	@FXML
	private Button cancelBtn;

	@FXML
	private Button checkBtn;

	@FXML
	private ProgressBar progress;

	@FXML
	private Tab filterTab;

	@FXML
	private TabPane tabPane;

	@FXML
	private Button loadBtn;

	@FXML
	private ImageView loadView;

	@FXML
	private Button saveBtn;

	@FXML
	private ImageView saveView;

	@FXML
	private Button selectBtn;

	@FXML
	private TextField schematronField;

	@FXML
	private Tab checksTab;

	@FXML
	private Label globalParametersLabel;

	@FXML
	private Label availableChecksLabel;

	@FXML
	private Label geometricChecksLabel;

	@FXML
	private Label semanticChecksLabel;

	@FXML
	private Label schematronFileLabel;

	@FXML
	private VBox zipArchiveOptionsVBox;
	@FXML
	private CheckBox streamCheckOption;
	@FXML
	private Label streamCheckLabel;
	@FXML
	private VBox streamCheckParametersVBox;
	@FXML
	private Label outputPathLabel;
	@FXML
	private TextField outputPathTextField;
	@FXML
	private Button outputPathBtn;
	@FXML
	private CheckBox pdfReportsOption;
	@FXML
	private CheckBox xmlReportsOption;

	private CityDoctorController controller;
	private FilterPane filterPane;
	private MainWindow window;
	private boolean zipArchiveMode;

	public CheckDialog(MainWindow window, Window parent, CityDoctorController controller) throws IOException {
		this.window = window;
		this.controller = controller;

		FXMLLoader loader = new FXMLLoader(CheckDialog.class.getResource("CheckDialog.fxml"));
		loader.setController(this);
		VBox box = loader.load();

		stage = new Stage();
		stage.getIcons().add(new Image(MainWindow.class.getResourceAsStream("icons/CityDoctor-Logo-rot_klein.jpg")));
		Scene scene = new Scene(box);
		stage.setScene(scene);
		stage.initOwner(parent);
		stage.initModality(Modality.APPLICATION_MODAL);
		stage.setTitle(Localization.getText("CheckDialog.title"));
	}

	public void initialize() {
		setupZipArchiveOptions();

		createEnableColumn(geometricTable);
		createNameColumn(geometricTable);
		createValueColumn(geometricTable);
		createUnitColumn(geometricTable);

		createEnableColumn(semanticTable);
		createNameColumn(semanticTable);
		createValueColumn(semanticTable);
		createUnitColumn(semanticTable);

		checksTab.setText(Localization.getText("CheckDialog.checksTab"));
		filterTab.setText(Localization.getText("CheckDialog.filterTab"));
		globalParametersLabel.setText(Localization.getText("CheckDialog.globalParametersLabel"));
		availableChecksLabel.setText(Localization.getText("CheckDialog.availableChecksLabel"));
		geometricChecksLabel.setText(Localization.getText("CheckDialog.geometricChecksLabel"));
		semanticChecksLabel.setText(Localization.getText("CheckDialog.semanticChecksLabel"));
		schematronFileLabel.setText(Localization.getText("CheckDialog.schematronFileLabel"));
		selectBtn.setText(Localization.getText("CheckDialog.selectBtn"));
		checkBtn.setText(Localization.getText("CheckDialog.checkBtn"));
		cancelBtn.setText(Localization.getText("CheckDialog.cancelBtn"));

		List<TableColumn<GlobalParameter, ?>> columns = globalParametersTable.getColumns();
		TableColumn<GlobalParameter, String> nameCol = new TableColumn<>(NAME);
		columns.add(nameCol);
		nameCol.setMinWidth(200);
		nameCol.setPrefWidth(325);
		nameCol.setCellValueFactory(new PropertyValueFactory<>("name"));

		setupValueColumns(columns);

		TableColumn<GlobalParameter, Unit> unitCol = new TableColumn<>(UNIT);
		columns.add(unitCol);
		unitCol.setMinWidth(50);
		unitCol.setPrefWidth(75);
		unitCol.setMaxWidth(100);
		unitCol.setEditable(false);
		unitCol.setCellValueFactory(new PropertyValueFactory<>("unit"));
		unitCol.setCellFactory(TextFieldTableCell.forTableColumn(new StringConverter<Unit>() {

			@Override
			public String toString(Unit object) {
				return object.getRepresentation();
			}

			@Override
			public Unit fromString(String string) {
				return Unit.valueOf(string);
			}
		}));

		loadImages();

		setupSelectButton();
		setupSaveButton();
		setupLoadButton();

		populateTreeWithChecks(geometricTable, RequirementType.GEOMETRY, true);
		populateTreeWithChecks(semanticTable, RequirementType.SEMANTIC, false);

		globalParametersTable.getItems().add(new GlobalParameter(NUMBER_OF_ROUNDING_PLACES, "8", Unit.NONE));
		globalParametersTable.getItems().add(new GlobalParameter("minVertexDistance", "0.0001", Unit.METER));

		cancelBtn.setOnAction(ea -> stage.close());
		setupCheckButton();

		tabPane.getSelectionModel().selectedIndexProperty().addListener((v, old, newV) -> {
			if (newV.intValue() == 1 && filterPane == null) {
				try {
					getFilterPane();
				} catch (IOException e) {
					window.showExceptionDialog(e);
				}
			}
		});
		
		schematronField.setText(ValidationConfiguration.CHECK_FOR_SOLID_XML);
	}

	private void setupValueColumns(List<TableColumn<GlobalParameter, ?>> columns) {
		TableColumn<GlobalParameter, String> valueCol = new TableColumn<>(VALUE);
		columns.add(valueCol);
		valueCol.setMinWidth(150);
		valueCol.setPrefWidth(200);
		valueCol.setEditable(true);
		valueCol.setCellValueFactory(new PropertyValueFactory<>("value"));
		valueCol.setCellFactory(column -> new TableEditCell<GlobalParameter, String>(new DefaultStringConverter()) {
			@Override
			public void updateItem(String item, boolean empty) {
				super.updateItem(item, empty);
				if (!empty && item != null) {
					GlobalParameter row = (GlobalParameter) getTableRow().getItem();
					if (row == null) {
						return;
					}
					row.setValue(item);
					if (NUMBER_OF_ROUNDING_PLACES.equals(row.getName())) {
						setEditable(false);
					}
				}
			}
		});
	}

	private void setupZipArchiveOptions() {
		streamCheckOption.selectedProperty().addListener(new ChangeListener<Boolean>() {
			@Override
			public void changed(ObservableValue<? extends Boolean> observable, Boolean oldValue, Boolean newValue) {
				streamCheckParametersVBox.setDisable(!streamCheckOption.isSelected());
			}
		});

		outputPathBtn.setOnAction(e -> {
			DirectoryChooser directoryChooser = new DirectoryChooser();
			directoryChooser.setInitialDirectory(Path.of(Settings.get(Settings.LAST_OPEN_FOLDER)).getParent().toFile());
			File f = directoryChooser.showDialog(stage.getOwner());
			if (f != null) {
				outputPathTextField.setText(f.getAbsolutePath());
				Settings.set(Settings.LAST_OPEN_FOLDER, f.getAbsolutePath());
			}


		});

		streamCheckOption.setText(Localization.getText("CheckDialog.streamCheckOption"));
		streamCheckLabel.setTooltip(new Tooltip(Localization.getText("CheckDialog.streamCheckLabel")));
		outputPathLabel.setText(Localization.getText("CheckDialog.outputPathLabel"));
		outputPathBtn.setText(Localization.getText("CheckDialog.outputPathBtn"));
		pdfReportsOption.setText(Localization.getText("CheckDialog.pdfReportsOption"));
		xmlReportsOption.setText(Localization.getText("CheckDialog.xmlReportsOption"));

		showZipArchiveOptions(false);
	}

	/**
	 * Shows the option panel for ZIP-Archive checking and enables the ZIP-Archive checking mode for this dialog.
	 *
	 * @param enabled
	 */
	public void showZipArchiveOptions(boolean enabled) {
		zipArchiveOptionsVBox.setDisable(!enabled);
		zipArchiveOptionsVBox.setVisible(enabled);
		zipArchiveMode = enabled;
	}

	private void setupSelectButton() {
		selectBtn.setOnAction(ae -> {
			FileChooser chooser = new FileChooser();
			List<String> extensions = new ArrayList<>();
			extensions.add("*.xml");
			String schematronFiles = Localization.getText("CheckDialog.schematronFiles");
			chooser.getExtensionFilters().add(new ExtensionFilter(schematronFiles, extensions));
			File dir = new File(Settings.get(Settings.LAST_OPEN_FOLDER, ""));
			if (dir.exists() && dir.isDirectory()) {
				chooser.setInitialDirectory(dir);
			}
			String schematronChooserTitle = Localization.getText("CheckDialog.schematronChooserTitle");
			chooser.setTitle(schematronChooserTitle);
			File file = chooser.showOpenDialog(stage);
			if (file != null) {
				schematronField.setText(file.getAbsolutePath());
			}
		});
	}

	private void setupLoadButton() {
		loadBtn.setOnAction(ae -> {
			FileChooser fc = new FileChooser();
			String validationConfiguration = Localization.getText("CheckDialog.validationConfiguration");
			fc.getExtensionFilters().add(new ExtensionFilter(validationConfiguration, "*.yml"));
			File dir = new File(Settings.get(Settings.LAST_OPEN_FOLDER, ""));
			if (dir.exists() && dir.isDirectory()) {
				fc.setInitialDirectory(dir);
			} else {
				Settings.set(Settings.LAST_OPEN_FOLDER, "");
			}
			File f = fc.showOpenDialog(stage);
			if (f != null) {
				Settings.set(Settings.LAST_OPEN_FOLDER, f.getParent());
				try {
					ValidationConfiguration config = ValidationConfiguration.loadValidationConfig(f.getAbsolutePath());
					applyConfig(config);
				} catch (IOException e) {
					window.showExceptionDialog(e);
				}
			}

		});
	}

	private void setupSaveButton() {
		saveBtn.setOnAction(ae -> {
			ValidationConfiguration config = createConfig();
			FileChooser fc = new FileChooser();
			fc.getExtensionFilters()
					.add(new ExtensionFilter(Localization.getText("CheckDialog.validationConfiguration"), "*.yml"));
			File dir = new File(Settings.get(Settings.LAST_OPEN_FOLDER, ""));
			if (dir.exists() && dir.isDirectory()) {
				fc.setInitialDirectory(dir);
			} else {
				Settings.set(Settings.LAST_OPEN_FOLDER, "");
			}
			File f = fc.showSaveDialog(stage);
			if (f != null) {
				Settings.set(Settings.LAST_OPEN_FOLDER, f.getParent());
				try {
					config.saveAs(f);
				} catch (IOException e) {
					window.showExceptionDialog(e);
				}
			}
		});
	}

	private void applyConfig(ValidationConfiguration config) throws IOException {
		FilterConfiguration filter = config.getFilter();
		if (filter != null) {
			getFilterPane().applyFilterConfig(filter);
		}
		Map<String, RequirementConfiguration> checks = config.getRequirements();
		for (Entry<String, RequirementConfiguration> e : checks.entrySet()) {
			for (TreeItem<TreeRequirement> ti : geometricTable.getRoot().getChildren()) {
				applyCheckConfig(e, ti);
			}
		}
		globalParametersTable.getItems().clear();
		globalParametersTable.getItems().add(new GlobalParameter(GlobalParameters.NUMBER_OF_ROUNDING_PLACES,
				config.getNumberOfRoundingPlacesAsString(), Unit.NONE));
		globalParametersTable.getItems().add(new GlobalParameter(GlobalParameters.MIN_VERTEX_DISTANCE,
				config.getMinVertexDistanceAsString(), Unit.METER));
		schematronField.setText(config.getSchematronFilePath());
	}

	private FilterPane getFilterPane() throws IOException {
		if (filterPane == null) {
			filterPane = new FilterPane();
			Platform.runLater(() -> filterTab.setContent(filterPane.getPane()));
		}
		return filterPane;
	}

	private void applyCheckConfig(Entry<String, RequirementConfiguration> e, TreeItem<TreeRequirement> ti) {
		Requirement r = ti.getValue().getRequirement();
		RequirementConfiguration cConfig = e.getValue();
		if (r.getId().equals(e.getKey())) {
			ti.getValue().getEnabledProperty().set(cConfig.isEnabled());
			for (Entry<String, String> paramEntry : cConfig.getParameters().entrySet()) {
				String name = paramEntry.getKey();
				String value = paramEntry.getValue();
				for (TreeItem<TreeRequirement> configItem : ti.getChildren()) {
					TreeRequirement parameterCheck = configItem.getValue();
					if (name.equals(parameterCheck.getNameProperty().getValue())) {
						parameterCheck.setValue(value);
					}
				}
			}
		}
	}

	private boolean collectCheckConfiguration(ValidationConfiguration config) {
		collectCheckInformationFromTree(config, geometricTable);
		collectCheckInformationFromTree(config, semanticTable);
		config.setSchematronFilePathInGlobalParameters(schematronField.getText());
		config.setMinVertexDistanceInGlobalParameters(
				Double.parseDouble(globalParametersTable.getItems().get(1).getValue()));
		// check if everything is okay with the configuration
		config.validateConfiguration();
		return true;
	}

	private void collectCheckInformationFromTree(ValidationConfiguration config, TreeTableView<TreeRequirement> table) {
		for (TreeItem<TreeRequirement> ti : table.getRoot().getChildren()) {
			Requirement r = ti.getValue().getRequirement();
			if (ti.getValue().getEnabledProperty().getValue()) {
				RequirementConfiguration cc = config.getRequirements().get(r.getId());
				cc.setEnabled(ti.getValue().isEnabled());
				// collect parameters
				collectParameters(cc, ti);
			} else {
				config.getRequirements().get(r.getId()).setEnabled(false);
			}
		}
	}

	private void loadImages() {
		try {
			try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/openFolderIcon.png")) {
				Image img = new Image(inStream);
				loadView.setImage(img);
			}
			try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/save.png")) {
				Image img = new Image(inStream);
				saveView.setImage(img);
			}
		} catch (IOException e) {
			// ignore close exception
		}
	}

	private void setupCheckButton() {
		checkBtn.setOnAction(ea -> {
			ValidationConfiguration config = createConfig();
			checkBtn.setDisable(true);
			cancelBtn.setDisable(true);
			stage.setOnCloseRequest(Event::consume);

			Thread t = new Thread(() -> {
				try {
					if (logger.isInfoEnabled()) {
						String startingChecks = Localization.getText("CheckDialog.startingChecks");
						logger.info(startingChecks);
					}
					if (zipArchiveMode) {
						Path outputPath = null;
						if (streamCheckOption.isSelected()) {
							if (outputPathTextField.getText().isEmpty()) {
								Alert warning = new Alert(Alert.AlertType.ERROR);
								warning.setTitle("Error");
								warning.setContentText(Localization.getText("CheckDialog.noPathError"));
								warning.showAndWait();
								return;
							}
							outputPath = Path.of(outputPathTextField.getText());
							Platform.runLater(() -> progress.setProgress(ProgressIndicator.INDETERMINATE_PROGRESS));
						}
						controller.startZipFileChecks(config, progress::setProgress, streamCheckOption.isSelected(),
								outputPath, pdfReportsOption.isSelected(),
								xmlReportsOption.isSelected());
					} else {
						controller.startChecks(config, progress::setProgress);
					}

					if (logger.isInfoEnabled()) {
						String checksDone = Localization.getText("CheckDialog.checksDone");
						logger.info(checksDone);
					}
				} catch (Exception e) {
					if (logger.isErrorEnabled()) {
						String failedChecks = Localization.getText("CheckDialog.failedChecks");
						logger.error(failedChecks, e);
					}
					Platform.runLater(() -> {
						ExceptionDialog dialog = new ExceptionDialog();
						dialog.show(e);
					});
				} finally {
					Platform.runLater(() -> {
						checkBtn.setDisable(false);
						cancelBtn.setDisable(false);
						stage.setOnCloseRequest(null);
						stage.close();
					});
				}
			});
			t.start();
		});
	}

	private ValidationConfiguration createConfig() {
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		config.setMinVertexDistanceInGlobalParameters(
				Double.parseDouble(globalParametersTable.getItems().get(1).getValue()));
		if (filterPane != null) {
			config.setFilter(filterPane.getConfiguration());
		}
		collectCheckConfiguration(config);
		if (schematronField.getText().isEmpty()) {
			return config;
		}
		File f = new File(schematronField.getText());
		if (!f.exists() || f.isDirectory() && logger.isWarnEnabled()) {
			String schematronFileNotExisting = Localization.getText("CheckDialog.schematronFileNotExisting");
			logger.warn(schematronFileNotExisting);
		} else {
			config.setSchematronFilePathInGlobalParameters(schematronField.getText());
		}
		return config;
	}

	private void collectParameters(RequirementConfiguration cc, TreeItem<TreeRequirement> ti) {
		Map<String, String> parameters = new HashMap<>();
		for (TreeItem<TreeRequirement> paramTi : ti.getChildren()) {
			String name = paramTi.getValue().getNameProperty().getValue();
			String value = paramTi.getValue().getValueProperty().getValue();
			parameters.put(name, value);
		}
		cc.setParameters(parameters);
	}

	private void createEnableColumn(TreeTableView<TreeRequirement> view) {
		TreeTableColumn<TreeRequirement, Boolean> enableColumn = new TreeTableColumn<>(CHECK_ENABLED);
		view.getColumns().add(enableColumn);
		enableColumn.setCellValueFactory(param -> {
			TreeRequirement check = param.getValue().getValue();
			if (check != null) {
				return check.getEnabledProperty();
			}
			return null;
		});

		StringConverter<Boolean> converter = new StringConverter<>() {

			@Override
			public String toString(Boolean object) {
				return " ";
			}

			@Override
			public Boolean fromString(String string) {
				return Boolean.parseBoolean(string);
			}
		};

		enableColumn.setCellFactory(list -> new CheckBoxTreeTableCell<TreeRequirement, Boolean>(null, converter) {

			@Override
			public void updateItem(Boolean item, boolean empty) {
				super.updateItem(item, empty);
				if (empty || item == null) {
					setGraphic(null);
					setText(null);
				}
			}
		});
	}

	private void createNameColumn(TreeTableView<TreeRequirement> view) {
		TreeTableColumn<TreeRequirement, String> nameColumn = new TreeTableColumn<>(NAME);
		nameColumn.setEditable(false);
		view.getColumns().add(nameColumn);
		nameColumn.setCellValueFactory(param -> {
			if (param.getValue().getValue() != null) {
				return param.getValue().getValue().getNameProperty();
			}
			return null;
		});
	}

	private void createValueColumn(TreeTableView<TreeRequirement> view) {
		TreeTableColumn<TreeRequirement, String> valueColumn = new TreeTableColumn<>(VALUE);
		view.getColumns().add(valueColumn);
		valueColumn.setCellValueFactory(param -> {
			if (param.getValue().getValue() != null) {
				return param.getValue().getValue().getValueProperty();
			}
			return null;
		});
		valueColumn.setCellFactory(col -> new TreeEditCell<>(new DefaultStringConverter()));
		valueColumn.setPrefWidth(100);
	}

	private void createUnitColumn(TreeTableView<TreeRequirement> view) {
		TreeTableColumn<TreeRequirement, String> unitColumn = new TreeTableColumn<>(UNIT);
		unitColumn.setEditable(false);
		view.getColumns().add(unitColumn);
		unitColumn.setCellValueFactory(param -> {
			if (param.getValue().getValue() != null) {
				return param.getValue().getValue().getUnitProperty();
			}
			return null;
		});
	}

	private void populateTreeWithChecks(TreeTableView<TreeRequirement> view, RequirementType type, boolean enabled) {
		view.setRoot(new TreeItem<>());
		List<Requirement> requirements = new ArrayList<>(Checks.getAvailableRequirements().values());
		Collections.sort(requirements, (r1, r2) -> r1.getId().compareTo(r2.getId()));
		for (Requirement r : requirements) {
			if (r.getType() != type) {
				continue;
			}
			TreeRequirement treeCheck = new TreeRequirement(r);
			treeCheck.setEnabled(enabled);
			TreeItem<TreeRequirement> ti = new TreeItem<>(treeCheck);
			view.getRoot().getChildren().add(ti);
			for (DefaultParameter dp : r.getDefaultParameter()) {
				TreeItem<TreeRequirement> dpTi = new TreeItem<>(new TreeRequirement(dp));
				ti.getChildren().add(dpTi);
			}
			ti.setExpanded(true);
		}
	}


	public void show() {
		Platform.runLater(() -> {
			if (controller.getCurrentConfig() != null) {
				for (GlobalParameter parameter : globalParametersTable.getItems()) {
					if (parameter.getName().equals(NUMBER_OF_ROUNDING_PLACES)) {
						parameter.setValue("" + controller.getCurrentConfig().getNumberOfRoundingPlaces());
					}
				}
			}
			globalParametersTable.refresh();
			progress.setProgress(0d);
			stage.showAndWait();
		});
	}

}
